@@ -4,12 +4,14 @@ import functools
 import json
 import operator
 import pathlib
+import stat
 import subprocess
 import sys
 import time
 import traceback
 from io import StringIO
-from os import environ, getenv, path
+from os import chmod, environ, getenv, path
+from typing import Tuple
 from uuid import uuid4
 
 import colorama
@@ -17,7 +19,10 @@ import pkg_resources
 import requests
 import yaml
 
-from kubedev.utils import KubedevConfig, KubernetesTools, YamlMerger
+from kubedev.utils import (CouldNotPullImageInCIException, KubedevConfig,
+                           KubernetesTools, YamlMerger)
+from kubedev.utils.errors import (CouldNotPullImageInCIException,
+                                  MissingFieldException)
 
 colorama.init(autoreset=True)
 
@@ -53,6 +58,9 @@ class RealFileAccessor:
   def abspath(self, filepath):
     return path.abspath(filepath)
 
+  def chmod_add_group_read(self, path):
+    return chmod(path, stat.S_IRGRP | stat.S_IREAD | stat.S_IWRITE)
+
   def mkdirhier(self, path):
     return pathlib.Path(path).mkdir(parents=True, exist_ok=True)
 
@@ -449,18 +457,15 @@ class Kubedev:
       print(f'{colorama.Fore.YELLOW}  (ℹ) kubedev is about to deploy your helm chart {helm_chart}.')
       print(f'{colorama.Fore.YELLOW}  (ℹ) If you want to inspect what is going on in your cluster, use this command:')
       print()
-      print(f'  docker run -it --rm --network {docker_network} --volume "{normalized_kubeconfig}:/tmp/kube_config" bitnami/kubectl --kubeconfig /tmp/kube_config <your-commands>')
-      print()
-      command = [
-        "/bin/sh",
-        "-c",
-        " ".join([
+      dockerRunKubctlCommand =[
           'docker',
           'run',
           '-i',
           '--rm',
           '--network',
           docker_network,
+          '--user',
+          '0', # Unfortunately, we have to start this container as root, because it could otherwise not read --volume mounted files
           '--volume',
           f'{normalized_kubeconfig}:/tmp/kube_config',
           '--volume',
@@ -475,7 +480,13 @@ class Kubedev:
           '/tmp/kube_config'] + \
           (["--kube-context", kubeContext] if kubeContext != None else []) + \
           ['--set', f'KUBEDEV_TAG={tag}'] + \
-          envs['cmdline'])
+          envs['cmdline']
+      print(f'  {" ".join(dockerRunKubctlCommand)}')
+      print()
+      command = [
+        "/bin/sh",
+        "-c",
+        " ".join(dockerRunKubctlCommand)
         ]
     else:
       envs = KubedevConfig.get_helm_set_env_args(kubedev, env_accessor, variable_overrides=variable_overrides, cmdline_as_list = False)
@@ -709,9 +720,9 @@ class Kubedev:
       print('❌ Check failed')
     return result
 
-  def run(self, configFileName, container, env_accessor=RealEnvAccessor(), shell_executor=RealShellExecutor(), printer=RealPrinter(), file_accessor=RealFileAccessor()):
+  def run(self, configFileName, container, env_accessor=RealEnvAccessor(), shell_executor=RealShellExecutor(), printer=RealPrinter(), file_accessor=RealFileAccessor(), extra_args=[]):
     return self.run_from_config(
-        self._load_config(configFileName, file_accessor), container, env_accessor=env_accessor, printer=printer, file_accessor=file_accessor)
+        self._load_config(configFileName, file_accessor), container, env_accessor=env_accessor, printer=printer, file_accessor=file_accessor, extra_args=extra_args)
 
   def run_from_config(self,
                       kubedev,
@@ -720,7 +731,8 @@ class Kubedev:
                       shell_executor=RealShellExecutor(),
                       printer=RealPrinter(),
                       file_accessor=RealFileAccessor(),
-                      tag_generator=TagGenerator()):
+                      tag_generator=TagGenerator(),
+                      extra_args=[]):
     images = KubedevConfig.get_images(kubedev, env_accessor)
     if not container in images:
       raise KeyError(
@@ -730,21 +742,30 @@ class Kubedev:
       currentTag = tag_generator.tag()
       buildResult = self.build_from_config(
           kubedev, container, currentTag, file_accessor=None, shell_executor=shell_executor, env_accessor=env_accessor)
-      interactive_flags = "--tty " if shell_executor.is_tty() else ""
+      interactive_flags = ["--tty"] if shell_executor.is_tty() else []
 
       if buildResult != 0:
         return buildResult
       else:
         (runEnvArgs, extraEnvs) = KubedevConfig.get_docker_run_envs(image, env_accessor=env_accessor)
+        volumeArgs = KubedevConfig.get_docker_run_volumes_list(Kubedev._field_optional(image, 'volumes', dict()), file_accessor, shell_executor)
+        publishArgs = KubedevConfig.get_docker_run_ports(image)
 
         command = [
           '/bin/sh',
           '-c',
-          f"docker run --interactive {interactive_flags}--rm " +
-          KubedevConfig.get_docker_run_volumes(Kubedev._field_optional(image, 'volumes', dict()), file_accessor, shell_executor) +
-          KubedevConfig.get_docker_run_ports(image) +
-          runEnvArgs +
-          f"{image['imageNameTagless']}:{currentTag}"
+          " ".join([
+            "docker",
+            "run",
+            "--interactive"] + \
+            interactive_flags + \
+            ["--rm"] + \
+            volumeArgs + \
+            publishArgs + \
+            runEnvArgs + \
+            [f"{image['imageNameTagless']}:{currentTag}"] + \
+            extra_args
+          )
         ]
         return shell_executor.execute(command, envVars=extraEnvs, check=False)
 
@@ -755,6 +776,14 @@ class Kubedev:
     else:
       return []
 
+  @staticmethod
+  def _pull_image(shell_executor: object, image: str) -> bool:
+    try:
+        shell_executor.execute(["docker", "pull", image])
+    except subprocess.CalledProcessError as e:
+      raise CouldNotPullImageInCIException(image) from e
+    return True
+
   def _run_docker_detached(
     self,
     network: str,
@@ -767,7 +796,7 @@ class Kubedev:
     cmd: list,
     env_accessor: object,
     shell_executor: object,
-    file_accessor: object) -> (str, bool):
+    file_accessor: object) -> Tuple[str, bool]:
       """
       Starts a container in detached mode and returns it's ID.
 
@@ -775,8 +804,14 @@ class Kubedev:
       """
       (imageDef, image, requiredEnvs, isFromKubedev) = Kubedev._build_image(rawImage, images)
       filteredRequiredEnvs = sorted([env for env in requiredEnvs if not env in variables]) # Note: the sorted() is important, otherwise the order would be
+
       if isFromKubedev and env_accessor.getenv('CI') is None:
-        self.build_from_image(imageDef, "none", env_accessor, shell_executor)
+          self.build_from_image(imageDef, "none", env_accessor, shell_executor)
+
+      if env_accessor.getenv('CI') is not None:
+      # try to pull each image & log potential errors
+          Kubedev._pull_image(shell_executor, image) # Will raise CouldNotPullImageInCIException when fails
+
       requiredEnvForwards = functools.reduce(operator.concat, [["--env", f'{envName}="${{{envName}}}"'] for envName in filteredRequiredEnvs], []) if isFromKubedev else []
       return self._run_docker_detached_impl(
         network,
@@ -833,7 +868,7 @@ class Kubedev:
               return (dockerId, name, False)
 
   @staticmethod
-  def _build_image(name: str, images: dict) -> (str, bool):
+  def _build_image(name: str, images: dict) -> Tuple[dict, str, dict, bool]:
       if len(name) > 3 and name[0] == '{' and name[-1] == '}':
           appName = name[1:-1]
           if appName in images:
@@ -845,7 +880,7 @@ class Kubedev:
   @staticmethod
   def _field_required(obj: dict, field: str, objectName: str):
       if not field in obj:
-          raise Exception(f'The field {field} is required in {objectName}')
+          raise MissingFieldException(f'The field {field} is required in {objectName}')
       else:
           return obj[field]
 
@@ -936,13 +971,15 @@ class Kubedev:
           ] + functools.reduce(operator.concat, [["--build-args", f'{arg}="{value}"'] for arg, value in buildArgs.items()], []) + [containerDir])]
       if shell_executor.execute(cmdBuild, envVars=buildArgs, check=False) != 0:
           return False
+
       # Step #1.1: For CronJobs: Build and push all apps
       if appType == 'cronjobs':
-        for appName, app in apps.items():
-          if self.build_from_config(kubedev, appName, uuid, file_accessor, shell_executor, env_accessor) != 0:
-            return False
-          if self._push_image(f"{images[appName]['imageNameTagless']}:{uuid}", shell_executor) != 0:
-            return False
+          for appName, app in apps.items():
+              if self.build_from_config(kubedev, appName, uuid, file_accessor, shell_executor, env_accessor) != 0:
+                  return False
+              if self._push_image(f"{images[appName]['imageNameTagless']}:{uuid}", shell_executor) != 0:
+                  return False
+
 
       # Step #2: Create the docker network, or spin up a Kind cluster:
       if appType == 'cronjobs':
@@ -982,6 +1019,8 @@ class Kubedev:
             # Spin up a new one-node Kubernetes-in-Docker cluster via `kind` (https://kind.sigs.k8s.io/)
             if shell_executor.execute(cmdKindCreateCluster, additionalVars) != 0:
               return False
+            # Add group-read permissions to the temporary kube-config, otherwise access may fail in containers running as non-root:
+            file_accessor.chmod_add_group_read(clusterConfig)
             clusterConfigContent = Kubedev._rewrite_kubeconfig_to(clusterConfig, f"{clusterName}-control-plane", 6443, file_accessor=file_accessor)
 
           wslClusterConfig = KubedevConfig.wsl_normalize(clusterConfig, file_accessor, shell_executor)
@@ -1012,7 +1051,15 @@ class Kubedev:
                   env_accessor,
                   shell_executor,
                   file_accessor) for serviceKey, service in self._field_optional(systemTestDefinition, 'services', dict()).items()]
+
           if appType == 'cronjobs':
+            # To make sure that the required images have been built in a previous CI job,
+            # we try to pull the images required in the helm chart to create
+            # more helpful error messages:
+            if env_accessor.getenv('CI') is not None:
+              for image in [image['imageName'] for _, image in images.items() if image['appType'] == 'deployments']:
+                Kubedev._pull_image(shell_executor, image) # Will raise CouldNotPullImageInCIException when fails
+
             runCronJobApiKey = tag_generator.tag()
             # Add variables for the test container to access the kubedev-run-cronjob-api service:
             variables = {
@@ -1070,12 +1117,16 @@ class Kubedev:
                 '--rm',
                 '--network',
                 network,
+                '--user',
+                '0',
                 '--volume',
                 f'{wslClusterConfig}:/tmp/kube_config',
                 'alpine/helm:2.16.9',
                 '--kubeconfig',
                 '/tmp/kube_config',
                 'init',
+                '--stable-repo-url',
+                'https://charts.helm.sh/stable',
                 '--service-account',
                 'tiller'
               ]
@@ -1142,8 +1193,6 @@ class Kubedev:
               print()
               result = False
           return result
-      except:
-          traceback.print_exc()
       finally:
           for startedContainer in startedContainers:
               containerId = startedContainer[0]
@@ -1176,15 +1225,12 @@ class Kubedev:
             print(f'{colorama.Fore.YELLOW}(ℹ) Keeping network {network} and cluster {reuseCluster} alive for further re-use.')
             print()
             print(f'Run this command to access your cluster:')
-            print(f'  docker run -it --rm --network {network} --volume "{wslClusterConfig}:/tmp/kube_config" bitnami/kubectl --kubeconfig /tmp/kube_config <your-commands>')
-
+            print(f'  docker run -it --rm --network {network} --user 0 --volume "{wslClusterConfig}:/tmp/kube_config" bitnami/kubectl --kubeconfig /tmp/kube_config <your-commands>')
           if result:
               print()
               print(f'{colorama.Fore.GREEN}System tests succeeded! 🌟🎉🥳')
               print()
-              return 0
           else:
               print()
               print(f'{colorama.Fore.RED}System tests failed! The logs of the services and the system test have been printed above.')
               print()
-              return 1
