@@ -151,6 +151,7 @@ class KubedevConfig:
           images[deploymentKey] = {
               "imageName": f"{imageRegistry}/{finalDeploymentName}:{tag}",
               "imageNameTagless": f"{imageRegistry}/{finalDeploymentName}",
+              "appType": appType,
               "appName": finalDeploymentName,
               "buildPath": KubedevConfig.get_buildpath(name, deploymentKey),
               "ports": deployment['ports'] if 'ports' in deployment else dict(),
@@ -200,13 +201,15 @@ class KubedevConfig:
   @staticmethod
   def get_docker_run_envs(image: dict, env_accessor: object) -> tuple:
     """
-    Returns a string with all "--env ..." parameters to the "docker run ..." call.
+    Returns a tuple with two entries:
+    [0]: A list with all ["--env", "<name>=<value>"] parameters to the "docker run ..." call.
+    [1]: A dict of extra environment variables and their values that must be passed to the container.
 
     :param image: One entry returned from KubedevConfig.get_images()
     """
     (envs, extraEnvs) = KubedevConfig.prepare_envs(image['containerEnvs'], env_accessor)
     return (
-      " ".join([f'--env {name}="{attribs["targetName"]}"' for name, attribs in sorted(envs.items())]) + " ",
+      [arg for arglist in [['--env', f'{name}="{attribs["targetName"]}"'] for name, attribs in sorted(envs.items())] for arg in arglist],
       extraEnvs
     )
 
@@ -252,21 +255,16 @@ class KubedevConfig:
 
     return list(itertools.chain(*[["--volume", get_path(hostPath, containerPath)] for hostPath, containerPath in volumes.items()]))
 
-  @staticmethod
-  def get_docker_run_volumes(volumes: dict, file_accessor, shell_executor) -> str:
-    args = KubedevConfig.get_docker_run_volumes_list(volumes, file_accessor, shell_executor)
-    return " ".join(args) + (" " if len(args) > 0 else "")
-
   @staticmethod
   def get_docker_run_ports(image):
     """
-    Returns a string with all "--publish ..." parameters to the "docker run ..." call.
+    Returns a list with all ["--publish", "<dev>:<container>"] port mappings parameters to the "docker run ..." call.
 
     :param image: One entry returned from KubedevConfig.get_images()
     """
-    return " ".join([
-      f"--publish {port['dev']}:{port['container']}" for port in image['ports'].values() if "container" in port and "dev" in port
-    ]) + " "
+    return [arg for argslist in [
+      ["--publish", f"{port['dev']}:{port['container']}"] for port in image['ports'].values() if "container" in port and "dev" in port
+    ] for arg in argslist]
 
   @staticmethod
   def get_helm_release_name(kubedev):
